<?php

namespace Drupal\ptools_hook;

use Drupal\ptools\SingletonTrait;
use Drupal\ptools\Utils\Text;

/**
 * Hook implementations handler class.
 */
class HookImplementationsHandler {

  use SingletonTrait;

  /**
   * The handler configuration.
   *
   * @var array
   */
  protected $config;

  /**
   * HookImplementationsHandler constructor.
   *
   * @param array $config
   *   The handler configuration.
   */
  public function __construct(array $config) {
    $this->config = $config;
  }

  /**
   * {@inheritdoc}
   */
  public static function create() {
    return new static(
      [
        'private_key' => drupal_get_private_key(),
        'hash_salt' => drupal_get_hash_salt(),
        'allow_runtime_implementations' => variable_get('ptools_hook_allow_runtime_implementations', TRUE),
      ]
    );
  }

  /**
   * Defines hook implementation at runtime.
   *
   * This is useful when not using generated hook implementations or when
   * rebuilding hook implementations on a cache clear.
   *
   * @param array|null $listeners_info
   *   (optional) If a listener information array is specified, the hook
   *   implementation cache will be rebuilt using that information. Defaults to
   *   to the cached hook implementations
   */
  public function defineRuntimeImplementations(array $listeners_info = NULL) {
    if (!$this->config['allow_runtime_implementations']) {
      return;
    }

    $cid = 'ptools_hook_implementations';

    $cache = cache_get($cid);
    if ($listeners_info === NULL && isset($cache->data)) {
      // Check cache data integrity before evaluating it, to prevent cache
      // tampering from resulting in arbitrary code execution.
      if ($this->getHash($cache->data['hook_implementations']) === $cache->data['hash']) {
        $hook_implementations = $cache->data['hook_implementations'];
      }
      else {
        watchdog('ptools_hook', 'Hook implementation cache was corrupt: <pre>@data</pre>', ['@data' => print_r($cache->data, TRUE)], WATCHDOG_ALERT);
        return;
      }
    }
    else {
      $hook_implementations = $this->dumpImplementations($listeners_info);
      $data = [
        'hook_implementations' => $hook_implementations,
        'hash' => $this->getHash($hook_implementations),
      ];
      cache_set($cid, $data);
    }

    eval($hook_implementations);
  }

  /**
   * Computes a hash for the specified hook implementations.
   *
   * @param string $hook_implementations
   *   The PHP code defining hook implementations.
   *
   * @return string
   *   A hash string.
   */
  protected function getHash($hook_implementations) {
    return hash_hmac('sha256', $hook_implementations, $this->config['private_key'] . $this->config['hash_salt']);
  }

  /**
   * Dumps the specified hook listener information as PHP code.
   *
   * @param array|null $listeners_info
   *   (optional) The listener information to be dumped. If none is specified,
   *   it is retrieved from scratch.
   *
   * @return string
   *   The dumped PHP code.
   */
  public function dumpImplementations(array $listeners_info = NULL) {
    $hook_implementations = '';

    if ($listeners_info === NULL) {
      $listeners_info = $this->getListenersInfo();
    }

    foreach ($listeners_info as $module => $module_info) {
      foreach ($module_info as $hook => $hook_info) {
        $single = count($hook_info['listeners']) === 1;
        $names = array_keys($hook_info['parameters']);
        $parameters = implode(', ', array_map(function ($name) use ($hook_info) { return ($hook_info['parameters'][$name] ? '&' : '') . "\${$name}"; }, $names));
        $invocation_parameters = implode(', ', array_map(function ($name) { return "\${$name}"; }, $names));
        $implementation_body = $single ? "" : "\$result = NULL;\n";

        foreach ($hook_info['listeners'] as $class => $info) {
          $instantiation = $info['singleton'] ? "{$class}::get()" : "new {$class}()";

          if ($single) {
            $implementation_body .=
   "\$class = '{$class}';
    if (class_exists(\$class) && method_exists(\$class, '{$info['method']}')) {
      \$listener = {$instantiation};
      return \$listener->{$info['method']}({$invocation_parameters});
    }
    else {
      ptools_hook_log_error('{$hook}', \$class);
    }";
          }
          else {
            $implementation_body .= "
    \$class = '{$class}';
    if (class_exists(\$class) && method_exists(\$class, '{$info['method']}')) {
      \$listener = {$instantiation};
      \$listener_result = \$listener->{$info['method']}({$invocation_parameters});
      if (is_array(\$listener_result)) {
        \$result = array_merge(\$result ?: [], \$listener_result);
      }
    }
    else {
      ptools_hook_log_error('{$hook}', \$class);
    }
";
          }
        }

        $return = $single ? "NULL" : "\$result";

        $hook_implementations .= "
if (!function_exists('{$module}_{$hook}')) {
  function {$module}_{$hook}({$parameters}) {
    {$implementation_body}
    return {$return};
  }
}
";
      }
    }

    return $hook_implementations;
  }

  /**
   * Discovers hook listeners and returns the related information.
   *
   * @return array
   *   An array of associative arrays keyed by module and hook names. Each leaf
   *   array contains information about a hook listener.
   */
  public function getListenersInfo() {
    $info = [];

    if (!function_exists('autoload')) {
      return [];
    }

    $classes = array_keys(iterator_to_array(autoload()));
    $listener_classes = preg_grep('/\\\\?Drupal\\\\[^\\\\]+\\\\Hook\\\\[^\\\\]+/', $classes);

    foreach ($listener_classes as $listener_class) {
      if (is_a($listener_class, HookListenerInterface::class, TRUE)) {
        $parts = explode('\\', $listener_class);

        if (isset($parts[1]) && module_exists($module = $parts[1])) {
          foreach (get_class_methods($listener_class) as $method) {
            $matches = [];
            if (preg_match('/^on([A-Z]\w+)$/', $method, $matches)) {
              $hook = Text::toSnakeCase($matches[1]);

              if (!isset($info[$module][$hook]['parameters'])) {
                $info[$module][$hook]['parameters'] = [];
              }

              $info[$module][$hook]['listeners'][$listener_class] = [
                'method' => $method,
                'singleton' => $this->isSingletonClass($listener_class),
              ];

              // We need to find out whether parameters are passed by reference
              // to be able to preserve references. If we have multiple
              // listeners for the same hook implementation, we need to merge
              // the reference information.
              try {
                $reflector = new \ReflectionMethod($listener_class, $method);
                foreach ($reflector->getParameters() as $parameter) {
                  $param_info = &$info[$module][$hook]['parameters'][$parameter->getName()];
                  $param_info = $param_info || $parameter->isPassedByReference();
                }
              }
              catch (\ReflectionException $e) {
                // This is not supposed to happen, as we just retrieved existing
                // class names.
                watchdog_exception('ptools', $e, NULL, [], WATCHDOG_ALERT);
              }
            }
          }
        }
      }
    }

    // Sort class information to make the invocation order predictable.
    foreach ($info as $module => &$module_info) {
      foreach ($module_info as $hook => &$hook_info) {
        ksort($hook_info['listeners']);
      }
    }

    return $info;
  }

  /**
   * Checks whether a class is a singleton.
   *
   * @param string $class
   *   The class name.
   *
   * @return bool
   *   TRUE if the class is a singleton, FALSE otherwise.
   */
  protected function isSingletonClass($class) {
    $is_singleton = function ($class) {
      return in_array(SingletonTrait::class, class_uses($class, FALSE));
    };

    if ($is_singleton($class)) {
      return TRUE;
    }

    foreach (class_parents($class, FALSE) as $parent_class) {
      if ($is_singleton($parent_class)) {
        return TRUE;
      }
    }

    return FALSE;
  }

}
