<?php

namespace Drupal\ptools;

/**
 * Trait useful to wrap database query objects to provide type hinting.
 */
trait DatabaseConnectionTrait {

  /**
   * The database connection.
   *
   * @var \DatabaseConnection
   */
  protected $database;

  /**
   * Instantiates a SELECT query object.
   *
   * @param $table
   *   The base table for this query.
   * @param $alias
   *   The alias of the base table of this query.
   * @param $options
   *   An array of options on the query.
   *
   * @return \SelectQuery
   *   An appropriate SelectQuery object for this database connection.
   *
   * @see db_select()
   */
  protected function getSelectQuery($table, $alias = NULL, array $options = []) {
    /** @var \SelectQuery $select */
    $select = $this->database->select($table, $alias, $options);
    return $select;
  }

  /**
   * Prepares and returns a DELETE query object.
   *
   * @param string $table
   *   The name of the table to operate on.
   * @param array $options
   *   An array of options on the query.
   *
   * @return \DeleteQuery
   *   A new DeleteQuery object.
   *
   * @see db_delete()
   */
  protected function getDeleteQuery($table, array $options = []) {
    return $this->database->delete($table, $options);
  }

  /**
   * Executes a query.
   *
   * @param string $query
   *   The prepared statement query to run.
   * @param array $args
   *   (optional) An array of values to substitute into the query.
   * @param array $options
   *   (optional) An array of options to control how the query operates.
   *
   * @return \DatabaseStatementInterface
   *   A prepared statement object, already executed.
   *
   * @see db_query()
   */
  protected function getQuery($query, $args = [], $options = []) {
    /** @var \DatabaseStatementInterface $statement */
    $statement = $this->database->query($query, $args, $options);
    return $statement;
  }

  /**
   * {@inheritdoc}
   */
  public function __sleep() {
    $properties = [];
    foreach ($this as $property => $value) {
      if ($property !== 'database') {
        $properties[] = $property;
      }
    }
    return $properties;
  }

}
