<?php

namespace Drupal\ptools\Drupal;

use Drupal\ptools\SingletonTrait;

/**
 * Base class for requirements providers.
 *
 * @see hook_requirements()
 */
abstract class RequirementsProviderBase {

  use SingletonTrait;

  /**
   * The translation function to be used.
   *
   * @var callable
   */
  protected $tCallback;

  /**
   * RequirementsProvider constructor.
   *
   * @param callable $t_callback
   *   The translation callback function to be used.
   */
  public function __construct(callable $t_callback) {
    $this->tCallback = $t_callback;
  }

  /**
   * {@inheritdoc}
   */
  public static function create() {
    return new static('get_t');
  }

  /**
   * Translates the specified string.
   *
   * @return string
   *   A translated string
   *
   * @see t()
   */
  protected function t() {
    $t = call_user_func($this->tCallback);
    return call_user_func_array($t, func_get_args());
  }

  /**
   * Returns requirements for the specified phase.
   *
   * @param string $phase
   *   The requirements phase.
   *
   * @return array
   *   An array of requirements as defined by hook_requirements().
   */
  public function getRequirements($phase) {
    switch ($phase) {
      case 'install':
        return $this->getInstallRequirements();

      case 'update':
        return $this->getUpdateRequirements();

      default:
        return $this->getRuntimeRequirements();
    }
  }

  /**
   * Returns install-time requirements.
   *
   * @return array
   *   An array of requirements as defined by hook_requirements().
   */
  public function getInstallRequirements() {
    return [];
  }

  /**
   * Returns update-time requirements.
   *
   * @return array
   *   An array of requirements as defined by hook_requirements().
   */
  public function getUpdateRequirements() {
    return [];
  }

  /**
   * Returns runtime requirements.
   *
   * @return array
   *   An array of requirements as defined by hook_requirements().
   */
  public function getRuntimeRequirements() {
    return [];
  }

}
