<?php

namespace Drupal\ptools\Entity;

use EntityDrupalWrapper;
use EntityListWrapper;
use EntityMetadataWrapperException;

/**
 * Entity metadata wrapper for an entity object.
 */
class EntityWrapper extends EntityDrupalWrapper {

  /**
   * Instantiates a new object based upon the entity or its ID.
   *
   * @param object|int $entity
   *   The entity object or its ID.
   *
   * @return static
   *   A new object based upon the called entity
   */
  public static function wrap($entity) {
    return new static($entity);
  }

  /**
   * Instantiates a new object based upon its values.
   *
   * @param array $values
   *   An associative array containing the entity values.
   *
   * @return static
   *   A new object based upon the called entity
   */
  public static function wrapValues(array $values) {
    return static::wrap((object) $values);
  }

  /**
   * Checks whether two wrappers are wrapping the same entity.
   *
   * @param \Drupal\ptools\Entity\EntityWrapper|null $a
   *   An entity wrapper object or NULL.
   * @param \Drupal\ptools\Entity\EntityWrapper|null $b
   *   An entity wrapper object or NULL.
   *
   * @return bool
   *   TRUE if the objects are wrapping the same entity, FALSE otherwise.
   */
  public static function isSameEntity(?EntityWrapper $a, ?EntityWrapper $b): bool {
    $id_a = $a ? (int) $a->getIdentifier() : NULL;
    $id_b = $b ? (int) $b->getIdentifier() : NULL;
    return $id_a === $id_b;
  }

  /**
   * Returns a wrapped cloned entity.
   *
   * Note that the wrapper itself is not cloned, so its state is not preserved.
   *
   * @return static
   *   A new entity wrapper object wrapping a clone of the current entity
   *   object.
   */
  public function cloneEntity() {
    return new static(clone $this->value());
  }

  /**
   * Returns the original identifier, in case the wrapped entity is invalid.
   *
   * @return string|null
   *   The entity ID originally specified.
   */
  public function getOriginalId(): ?string {
    return $this->id;
  }

  /**
   * Checks whether the entity is new.
   *
   * @retrun bool
   *   TRUE if the entity is new, FALSE otherwise.
   */
  public function isNew() {
    $entity = $this->value();
    if (!empty($entity->is_new)) {
      return TRUE;
    }
    $id = $this->getIdentifier();
    return !$id;
  }

  /**
   * Renders a view mode for the given entity.
   *
   * @param $view_mode
   *   A view mode as used by this entity type, e.g. 'full', 'teaser'...
   * @param $langcode
   *   (optional) A language code to use for rendering. Defaults to the global
   *   content language of the current request.
   * @param $page
   *   (optional) If set will control if the entity is rendered: if TRUE
   *   the entity will be rendered without its title, so that it can be embedded
   *   in another context. If FALSE the entity will be displayed with its title
   *   in a mode suitable for lists.
   *   If unset, the page mode will be enabled if the current path is the URI
   *   of the entity, as returned by entity_uri().
   *   This parameter is only supported for entities which controller is a
   *   EntityAPIControllerInterface.
   *
   * @return string
   *   The rendered content.
   */
  public function render($view_mode = 'full', $langcode = NULL, $page = NULL) {
    $build = $this->view($view_mode, $langcode, $page);
    return render($build);
  }

  /**
   * Safely gets an item list.
   *
   * @param string $field_name
   *   The field name.
   *
   * @return \EntityListWrapper
   *   A list of field items.
   */
  protected function safeGetItems($field_name) {
    try {
      return $this->get($field_name);
    }
    catch (EntityMetadataWrapperException $e) {
      return new EntityListWrapper('', []);
    }
  }

  /**
   * Safely gets a field value.
   *
   * @param string $field_name
   *   The field name.
   *
   * @return mixed
   *   The field value or NULL if not available.
   */
  protected function safeGetValue($field_name) {
    /** @var \EntityMetadataWrapper $item */
    try {
      $item = $this->get($field_name);
      return $item->value();
    }
    catch (EntityMetadataWrapperException $e) {
      return NULL;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function __get($name) {
    $entity = $this->value();
    if (isset($entity->{$name}) && is_scalar($entity->{$name})) {
      $get = $entity->{$name};
      return $get;
    }
    return parent::__get($name);
  }

  /**
   * Does the field have a value?
   *
   * @param string $field_name
   *   The name of the string
   *
   * @return bool
   */
  public function hasValue($field_name) {
    $value = $this->safeGetValue($field_name);
    return isset($value);
  }

  /**
   * Checks whether the entity has changes to the specified field.
   *
   * @param string $field_name
   *   The field name.
   * @param string $property_name
   *   (optional) The name of the property to check. By default all properties
   *   are checked.
   * @param bool $ignore_order
   *   (optional) Whether item order should be taken into account. Defaults to
   *   FALSE.
   * @param string $langcode
   *   (optional) The field language to be used. Defaults to the wrapper's
   *   active one.
   *
   * @return bool
   *   TRUE if the entity has field changes, FALSE otherwise.
   */
  public function hasFieldChanges($field_name, $property_name = NULL, $ignore_order = FALSE, $langcode = NULL) {
    if ($this->isNew()) {
      return FALSE;
    }

    if (!isset($this->data->original)) {
      $this->data->original = entity_load_unchanged($this->type, $this->getIdentifier());
    }
    if (!isset($langcode)) {
      $langcode = $this->getPropertyLanguage() ?: LANGUAGE_NONE;
    }

    $entity = $this->data;
    $original = $this->data->original;

    if (!empty($entity->{$field_name}[$langcode]) && empty($original->{$field_name}[$langcode])) {
      return TRUE;
    }
    elseif (empty($entity->{$field_name}[$langcode]) && !empty($original->{$field_name}[$langcode])) {
      return TRUE;
    }
    elseif (empty($entity->{$field_name}[$langcode]) && empty($original->{$field_name}[$langcode])) {
      return FALSE;
    }

    if ($property_name) {
      $callback = function ($item) use ($property_name) {
        return (int) $item[$property_name];
      };
      $new_items = array_map($callback, $entity->{$field_name}[$langcode]);
      $old_items = array_map($callback, $original->{$field_name}[$langcode]);
    }
    else {
      $new_items = $entity->{$field_name}[$langcode];
      $old_items = $original->{$field_name}[$langcode];
    }

    // We do not care whether items were reordered.
    if ($ignore_order) {
      asort($new_items);
      asort($old_items);
    }

    return $new_items !== $old_items;
  }

}
