<?php

namespace Drupal\ptools\Entity;

/**
 * Base class for taxonomy term wrappers.
 */
abstract class TaxonomyTermBundleWrapperBase extends EntityBundleWrapperBase {

  /**
   * {@inheritdoc}
   */
  public static function wrapValues(array $values) {
    $values += [
      'parent' => [0],
      'description' => '',
    ];
    if (!isset($values['vid'])) {
      $vocabulary = taxonomy_vocabulary_machine_name_load(static::getBundleName());
      $values['vid'] = $vocabulary->vid;
    }
    return parent::wrapValues($values);
  }

  /**
   * {@inheritdoc}
   */
  protected static function getEntityType(): string {
    return 'taxonomy_term';
  }

  /**
   * Checks whether the wrapped term is valid.
   *
   * @return bool
   *   TRUE if the term is valid, FALSE otherwise.
   */
  public function isValidTerm(): bool {
    return $this->getBundle() === static::getBundleName();
  }

  /**
   * Returns the taxonomy item name.
   *
   * @return string
   *   The item name.
   */
  public function getName(): string {
    return (string) $this->safeGetValue('name');
  }

  /**
   * Sets the taxonomy item name.
   *
   * @param string $name
   *   The item name.
   */
  public function setName(string $name): void {
    $this->safeGetItems('name')->set($name);
  }

  /**
   * Returns the taxonomy item description.
   *
   * @return string
   *   The item description.
   */
  public function getFormattedDescription(): string {
    return (string) $this->safeGetValue('description');
  }

  /**
   * Returns the raw taxonomy item description.
   *
   * @return string
   *   The raw item description.
   */
  public function getDescription(): string {
    return (string) $this->safeGetItems('description')->raw();
  }

  /**
   * Set the taxonomy item description.
   *
   * @param string $description
   *   The item description.
   */
  public function setDescription(string $description): void {
    $this->safeGetItems('description')->set($description);
  }

  /**
   * Returns the taxonomy item description format.
   *
   * @return string
   *   The item description format.
   */
  public function getFormat(): ?string {
    return $this->value()->format ?? NULL;
  }

  /**
   * Set the taxonomy item description format.
   *
   * @param string|NULL $format
   *   The item description format.
   */
  public function setFormat(?string $format): void {
    $this->value()->format = $format;
  }

  /**
   * Returns the taxonomy item weight.
   *
   * @return int
   *   The item weight.
   */
  public function getWeight(): int {
    return (int) $this->safeGetValue('weight');
  }

  /**
   * Sets the taxonomy item weight.
   *
   * @param int $weight
   *   The item weight.
   */
  public function setWeight(int $weight): void {
    $this->safeGetItems('weight')->set($weight);
  }

  /**
   * Return the term parents.
   *
   * @return static[]
   *   An associative array of parent term wrappers keyed by parent ID.
   */
  public function getParents(): array {
    $parents = [];
    $term = $this->value();
    $properties = ['parents', 'parent'];

    while (($property = array_shift($properties)) && !$parents) {
      if (isset($term->{$property})) {
        $pids = $term->{$property};
        // At times the "parent" property may contain a single parent ID.
        if (is_scalar($pids)) {
          $pids = [$pids];
        }
        foreach ($pids as $pid) {
          // Forms may put random cruft inside the "parents" property.
          if (is_numeric($pid) && $pid) {
            $parents[$pid] = static::wrap($pid);
          }
        }
      }
    }

    return $parents;
  }

  /**
   * Sets the term parents.
   *
   * @param static[] $parents
   *   An array of parent term wrappers.
   */
  public function setParents(array $parents): void {
    $ids = array_map(
      function (
        TaxonomyTermBundleWrapperBase $parent) {
          return $parent->getIdentifier();
        },
      $parents
    );
    $ids = $ids ?: [0];
    $this->safeGetItems('parent')->set($ids);
    $this->value()->parents = $ids;
  }

}
