<?php

namespace Drupal\ptools\Queue;

use Drupal\ptools\Plugin\PluginBase;

/**
 * Base class for queue handler plugins.
 */
abstract class QueueHandlerBase extends PluginBase implements QueueHandlerInterface {

  /**
   * Shutdown buffer useful to trigger garbage collection in case of OOMs.
   *
   * @var bool
   */
  protected static $shutdownBuffer;

  /**
   * The queue name.
   *
   * @var string
   */
  protected $queueName;

  /**
   * The handled queue.
   *
   * @var \DrupalQueueInterface
   */
  protected $queue;

  /**
   * The item currently being processed.
   *
   * @var \stdClass
   */
  protected $currentItem;

  /**
   * The item processing start time in microseconds.
   *
   * @var float
   */
  protected $itemStart;

  /**
   * The item processing end time in microseconds.
   *
   * @var float
   */
  protected $itemEnd;

  /**
   * QueueHandlerBase constructor.
   *
   * @param array $definition
   *   The plugin definition.
   */
  public function __construct(array $definition) {
    parent::__construct($definition);

    // Register a shutdown function to handle fatal errors (e.g. out of memory
    // errors) gracefully.
    if (!static::$shutdownBuffer) {
      static::$shutdownBuffer = str_pad('*', 1024);
      $self = $this;
      drupal_register_shutdown_function(function () use ($self) {
        // Trigger garbage collection.
        $self::$shutdownBuffer = NULL;
        $error = error_get_last();
        if (isset($error)) {
          $variables = [
            '!message' => $error['message'],
            '!file' => $error['file'],
            '!line' => $error['line'],
          ];
          watchdog('ptools_queue', '!message (line !line of !file).', $variables, WATCHDOG_ERROR);
          $self->handleItemFailure($self->currentItem);
        }
      });
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getWorkerDefinition() {
    return [
      'worker callback' => [QueueHandlerFactory::class, $this->getName()],
      'time' => $this->definition['time'],
      'skip on cron' => $this->definition['skip_on_cron'],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function setQueueContext($queue_name, \DrupalQueueInterface $queue) {
    $this->queueName = $queue_name;
    $this->queue = $queue;
  }

  /**
   * {@inheritdoc}
   */
  public function queueItems(array $items = NULL) {
    if (isset($items)) {
      $processed = $this->createItems($items);
    }
    else {
      $processed = 0;
      $total = $this->getPendingItemsCount();
      while ($processed < $total) {
        $items = $this->fetchPendingItems($processed);
        $processed += $this->createItems($items);
        $this->log(format_string('-> @count / @total processed', ['@count' => $processed, '@total' => $total]));
      }
    }
    $this->log(format_string('@count items added to the queue.', ['@count' => $processed]));
  }

  /**
   * Returns the total number of pending items.
   *
   * @return int
   *   The pending items count.
   */
  protected function getPendingItemsCount() {
    return 0;
  }

  /**
   * Fetches a batch of pending items to be queued.
   *
   * @param int $offset
   *   (optional) The offset index to start from. Defaults to 0.
   *
   * @return mixed[]
   *   An array of pending items.
   */
  protected function fetchPendingItems($offset = 0) {
    return [];
  }

  /**
   * Creates the specified queue items.
   *
   * @param mixed[] $items
   *   An array of queue items.
   *
   * @return int
   *   The number of processed items.
   */
  protected function createItems(array $items) {
    $this->queue->createQueue();
    foreach ($items as $item) {
      $this->queue->createItem([
        'item' => $item,
        'attempts' => $this->definition['max_attempts'],
      ]);
    }
    return count($items);
  }

  /**
   * {@inheritdoc}
   */
  public function processItems($limit = -1, $time_limit = NULL) {
    $processed = 0;
    $failed = 0;
    $count = intval($limit) ?: -1;
    if (isset($time_limit)) {
      $end = time() + $time_limit;
    }

    while ($count-- !== 0 && (!isset($end) || time() < $end) && ($item = $this->queue->claimItem($this->definition['lease_time']))) {
      try {
        $this->handleItemStart($item);
        $this->processItem($item->data['item']);
        $this->handleItemSuccess($item);
        $processed++;
      }
      catch (\Exception $e) {
        // In case of exception log it and leave the item in the queue to be
        // processed again later.
        watchdog_exception('ptools_queue', $e);
        $this->handleItemFailure($item);
        if ($item->data['attempts'] <= 1) {
          $failed++;
        }
      }
    }

    $args = [
      '@processed_count' => $processed,
      '@failed_count' => $failed,
      '@total' => $this->queue->numberOfItems(),
    ];
    $this->log(format_string('@processed_count queue items processed, @failed_count failed, @total remaining.', $args));

    return $processed;
  }

  /**
   * Handles item processing start.
   *
   * @param \stdClass $item
   *   A queue item.
   */
  protected function handleItemStart($item) {
    $this->currentItem = $item;
    // If the queue is expected to be empty at the end of the process, we need
    // to delete the item before processing it, in case the process fails
    // without performing proper clean-up.
    if ($this->definition['clear_failed']) {
      $this->queue->deleteItem($item);
    }
    $this->itemStart = microtime(TRUE);
  }

  /**
   * Handles item processing success.
   *
   * @param \stdClass $item
   *   A queue item.
   */
  protected function handleItemSuccess($item) {
    $this->itemEnd = microtime(TRUE);
    $this->queue->deleteItem($item);
    $this->currentItem = NULL;
  }

  /**
   * Handles item processing failure.
   *
   * @param \stdClass $item
   *   A queue item.
   */
  protected function handleItemFailure($item) {
    $this->itemEnd = microtime(TRUE);

    // Before giving up, we queue up the item again until we reach the
    // maximum number of attempts.
    if ($item->data['attempts'] > 1) {
      $this->queue->deleteItem($item);
      $item->data['attempts']--;
      $this->queue->createItem($item->data);
    }
    else {
      $args = [
        '@name' => $this->queueName,
        '@item_id' => $item->item_id,
        '@count' => $this->definition['max_attempts'],
        '@item' => print_r($item->data['item'], TRUE),
      ];
      watchdog('ptools_queue', 'Item "@item_id" of queue "@name" could not be processed after @count attempts: <pre>@item</pre>', $args, WATCHDOG_ERROR);
    }

    $this->currentItem = NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function clearQueue() {
    $processed = $this->queue->numberOfItems();
    $this->queue->deleteQueue();
    $this->log(format_string('@count items removed from queue.', ['@count' => $processed]));
    return $processed;
  }

  /**
   * Logs a message.
   *
   * @param string $message
   *   A message string.
   */
  protected function log($message) {
    // @todo Inject logger instead.
    if (function_exists('drush_log')) {
      drush_log($message, 'ok');
    }
  }

}
