<?php

namespace Drupal\ptools;

/**
 * Trait useful to implement the singleton pattern.
 */
trait SingletonTrait {

  /**
   * The singleton instance.
   *
   * @var static
   */
  protected static $instance;

  /**
   * Returns the singleton instance.
   *
   * This method should be overridden when the constructor has arguments.
   *
   * @param array $context
   *   (optional) An associative array of additional dependencies. Defaults to
   *   none.
   *
   * @return static
   *   The singleton instance.
   */
  public static function get(array $context = []) {
    if (!isset(static::$instance)) {
      static::$instance = static::create($context);
    }
    return static::$instance;
  }

  /**
   * Actually instantiates the object.
   *
   * @param array $context
   *   (optional) An associative array of additional dependencies. Defaults to
   *   none.
   *
   * @return static
   *   An object instance.
   */
  public static function create(array $context = []) {
    // Allow singletons to inherit the factory method from their parent class.
    $parent_class = get_parent_class(static::class);
    if (method_exists($parent_class, 'create')) {
      if (!isset($context['class'])) {
        $context['class'] = static::class;
      }
      return $parent_class::create($context);
    }
    else {
      return new static();
    }
  }

  /**
   * Ensures that a new object will be instantiated the next time.
   */
  public static function reset() {
    static::$instance = NULL;
  }

}
