<?php

namespace Drupal\ptools\Utils;

/**
 * Math-related utility API.
 */
class Math {

  /**
   * The default compare scale value.
   *
   * @todo Figure out whether this value is always sensible.
   */
  const DEFAULT_COMPARE_SCALE = 6;

  /**
   * Rounds a decimal number to a precision making comparisons meaningful.
   *
   * @param float|double $number
   *   A decimal number.
   *
   * @return double
   *   A rounded double-precision number.
   */
  public static function doubleRound($number) {
    return (double) round((double) $number, static::DEFAULT_COMPARE_SCALE);
  }

  /**
   * Returns the smallest significant double precision number.
   *
   * @return double
   *   A double precision value.
   */
  public static function getMinDoublePrecision() {
    return (double) pow(10, -1 * static::DEFAULT_COMPARE_SCALE);
  }

  /**
   * Compares two arbitrary precision numbers.
   *
   * This function uses http://php.net/manual/en/function.bccomp.php if
   * available.
   *
   * @param mixed $first
   *   The first numeric value to compare.
   * @param mixed $second
   *   The second numeric value to compare.
   * @param int $scale
   *   (optional) The optional scale parameter is used to set the number of
   *   digits after the decimal place which will be used in the comparison.
   *   Defaults to ::DEFAULT_COMPARE_SCALE.
   *
   * @return int
   *   0 if the two values are equal, 1 if the first value is larger than the
   *   second, -1 otherwise.
   *
   * @see ::getMinDoublePrecision()
   */
  public static function doubleCompare($first, $second, $scale = NULL) {
    if (!isset($scale)) {
      $scale = static::DEFAULT_COMPARE_SCALE;
    }
    if (function_exists('bccomp')) {
      return bccomp($first, $second, $scale);
    }
    else {
      $diff = (double) $first - (double) $second;
      if (abs($diff) < static::getMinDoublePrecision()) {
        return 0;
      }
      else {
        return $diff / abs($diff);
      }
    }
  }

  /**
   * Returns a value rounded to a given precision, unless it would match one of
   * the targets, in which case round it to the next useful precision.
   *
   * @param float $value
   *   The value to round.
   * @param float[] $targets
   *   The targets to avoid when rounding.
   * @param int $min_precision
   *   The minimum required precision.
   * @param bool $percentage
   *   (optional) Whether the value and the targets are percentages in 0..1
   *   format. Defaults to FALSE.
   *
   * @return string
   *   The formatted rounded value.
   */
  public static function variablePrecisionRoundFormat($value, $targets, $min_precision, $percentage = false) {
    if (is_int($value)) {
      return (string) ($percentage ? $value * 100 : $value);
    }

    if ($percentage) {
      $value *= 100;
      $targets = array_map(function ($value) {return $value * 100;}, $targets);
    }

    $results = array();
    foreach ($targets as $target) {
      $precision = $min_precision;
      if (static::doubleCompare($value, $target) != 0) {
        do {
          $result = round($value, $precision++);
        } while (static::doubleCompare($result, $target) == 0 && $precision < 9);
        $precision--;
      }
      $results[] = $precision;
    }

    // Get the result with the highest precision.
    $precision = max($results);
    $result = round($value, $precision);
    // Make sure also .0 decimals are returned.
    return sprintf("%.{$precision}f", $result);
  }

  /**
   * Counts the values of a multidimensional array.
   *
   * This differs from count($data, COUNT_RECURSIVE) because that will also
   * count "nodes" in the array graph, while this will only count leaves.
   *
   * @param array $data
   *   An array.
   *
   * @return int
   *   The number of array values.
   */
  public static function countValues(array $data) {
    $count = 0;
    array_walk_recursive($data, function () use (&$count) { $count++; });
    return $count;
  }

}
