<?php

namespace Drupal\ptools\Utils;

use Drupal\ptools\DatabaseConnectionTrait;
use Drupal\ptools\SingletonTrait;

/**
 * Update function utility API (mostly SQL-specific).
 */
class Update {

  use SingletonTrait;
  use DatabaseConnectionTrait;

  /**
   * Update constructor.
   *
   * @param \DatabaseConnection $database
   *   The database connection.
   */
  public function __construct(\DatabaseConnection $database) {
    $this->database = $database;
  }

  /**
   * {@inheritdoc}
   */
  public static function create() {
    return new static(
      \Database::getConnection()
    );
  }

  /**
   * Returns the SQL tables the specified field is stored in.
   *
   * @param string $field_name
   *   A field name.
   *
   * @return string[]
   *   The field table names.
   */
  public function getFieldTables(string $field_name): array {
    $field = field_info_field($field_name);
    return [
      'data' => _field_sql_storage_tablename($field),
      'revision' => _field_sql_storage_revision_tablename($field),
    ];
  }

  /**
   * Changes the bundle of the specified entities.
   *
   * @param string $entity_type
   *   The entity type.
   * @param string $old_bundle
   *   The old bundle name.
   * @param string $new_bundle
   *   The new bundle name.
   * @param array $ids
   *   (optional) An array of entity IDs to process. Defaults to all entities
   *   of the old bundle.
   */
  public function changeBundle(string $entity_type, string $old_bundle, string $new_bundle, array $ids = []): void {
    $entity_info = entity_get_info($entity_type);
    $bundle_key = $entity_info['entity keys']['bundle'];
    $old_bundle_value = $old_bundle;
    $new_bundle_value = $new_bundle;

    if ($entity_type === 'taxonomy_term') {
      $bundle_key = 'vid';
      $old_vocabulary = taxonomy_vocabulary_machine_name_load($old_bundle);
      $old_bundle_value = $old_vocabulary->vid;
      $new_vocabulary = taxonomy_vocabulary_machine_name_load($new_bundle);
      $new_bundle_value = $new_vocabulary->vid;
    }

    $base_tables = [
      $entity_info['base table'],
      $entity_info['revision table'] ?? NULL,
    ];
    foreach (array_filter($base_tables) as $base_table) {
      db_update($base_table)
        ->fields([$bundle_key => $new_bundle_value])
        ->condition($bundle_key, $old_bundle_value)
        ->execute();
    }

    foreach (field_info_instances($entity_type, $old_bundle) as $field_name => $instance) {
      foreach ($this->getFieldTables($field_name) as $table_name) {
        $query = db_update($table_name)
          ->fields(['bundle' => $new_bundle])
          ->condition('bundle', $old_bundle);

        if ($ids) {
          $query->condition('entity_id', $ids);
        }

        $query->execute();
      }
    }

    entity_get_controller($entity_type)->resetCache($ids);
  }

  /**
   * Deletes a field stored in a SQL database.
   *
   * @param string $field_name
   *   A field name.
   */
  public function deleteField(string $field_name): void {
    $field = field_info_field($field_name);
    if (!$field) {
      return;
    }

    $table_names = [
      _field_sql_storage_tablename($field),
      _field_sql_storage_revision_tablename($field),
    ];

    foreach ($table_names as $table_name) {
      $this->database
        ->truncate($table_name)
        ->execute();
    }

    field_delete_field($field_name);
  }

  /**
   * Deletes the specified instances.
   *
   * @param string $entity_type
   *   The entity type.
   * @param string $bundle
   *   The bundle name.
   * @param string[] $field_names
   *   The field instance names.
   */
  public function deleteInstances($entity_type, $bundle, array $field_names): void {
    foreach ($field_names as $field_name) {
      $instance = field_info_instance($entity_type, $field_name, $bundle);
      if ($instance) {
        field_delete_instance($instance);
      }
    }
  }

  /**
   * Deletes all field records having NULL values for the specified property.
   *
   * This is SQL storage-specific.
   *
   * @param string $field_name
   *   The field name.
   * @param string $property_name
   * The field property name.
   */
  public function purgeNullFieldRecords($field_name, $property_name) {
    $field = field_info_field($field_name);
    $column_name = _field_sql_storage_columnname($field_name, $property_name);

    $table_names = [
      _field_sql_storage_tablename($field),
      _field_sql_storage_revision_tablename($field),
    ];

    foreach ($table_names as $table_name) {
      $this->getDeleteQuery($table_name)
        ->isNull($column_name)
        ->execute();
    }
  }

}
